(* solve the Minimum d-Bounded-Degree Deletion problem 
   Copyright (C) 2009 Hannes Moser, Manuel Sorge

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.  *)


(* let f_debug = ref (fun s -> ());; *)

let nt_kernelization               = ref true;;
let exhaustive_interleaving        = ref true;;
let buss_interleaving              = ref true;;
let deg_one_interleaving           = ref true;;
let deg_two_interleaving           = ref true;;
let degree_less_than_or_equal_d_interleaving = ref false;;
let cliquerule_interleaving        = ref true;;
let guided_branch_vertex_selection = ref true;;
let greedysol_branching            = ref true;;
let forbid_branch_vertices         = ref true;;
let edge_rule                      = ref true;;
let bound_rule                     = ref true;;
let improve_lower_bound            = ref true;;
let forbid_sol_subsets             = ref false;;
let branch_greedy_first            = ref false;;
let min_branch_candidates          = ref 10;;
let interleaving_interval          = ref 1;;
let greedyopt_interval             = ref 1;;

let lb_executions = ref 0;;
let lb_improvements = ref 0;;

(* TODO extra bounding functions that compute multiple bounds and choose 
   the best one. *)

let lower_bound_f g d =
  lb_executions := !lb_executions + 1;

  let rec pack_stars sp forbidden g =
    let max_deg = FGraph.max_deg g in

    let rec get_vertex deg =
      if deg > max_deg then raise Not_found
      else
	let deg_vertices = IntSet.minus (FGraph.deg_vertices g deg) forbidden in
	  if (IntSet.is_empty deg_vertices) then
            get_vertex (deg+1)
	  else
	    IntSet.choose deg_vertices
    in

    let pop_star g v leaves =
      let rec loop star n g leaves =
	if leaves = 0 then star, g
	else
	  let v, n = IntSet.pop n in
	    loop (IntSet.add star v) n (FGraph.delete_vertex g v) (leaves - 1)
      in
      let star, g = 
	loop (IntSet.singleton v) (FGraph.neighbors g v) (FGraph.delete_vertex g v) leaves
      in
	(v, star), g
    in

      if max_deg <= d then sp, g
      else
	let continue, sp, g =
	  try 
	    let star, g = pop_star g (get_vertex (d + 1)) (d + 1) in
	      true, (star :: sp), g
	  with Not_found -> false, sp, g
	in
	  if not continue then sp, g
	  else pack_stars sp IntSet.empty g
  in

  let rec opt_starpacking rest_g sp sp' =
    if sp = [] then sp'
    else if (FGraph.num_vertices rest_g) < (d + 1) * (List.length sp) then sp @ sp'

    else
      let (root, star') as star, sp = List.hd sp, List.tl sp in	
      let sub_g = 
	IntSet.fold begin fun rest_g v ->
	  IntSet.fold begin fun rest_g w -> 
	    try FGraph.connect rest_g v w 
	    with Not_found -> rest_g
	  end (FGraph.neighbors g v) (FGraph.add_vertex rest_g v)
	end star' rest_g
      in
      let sub_sp, sub_g = pack_stars sp (IntSet.singleton root) sub_g in
	if List.length sub_sp - List.length sp <= 1 then
	  opt_starpacking rest_g sp (star :: sp')
	else begin
	  lb_improvements := !lb_improvements + 1;
	  (* prerr_endline "Lower bound improved."; *)
	  opt_starpacking sub_g (sub_sp @ sp') []
	end
  in

  let sp, rest_g = pack_stars [] IntSet.empty g in
  let sp = 
    if !improve_lower_bound (*&& (FGraph.num_vertices rest_g) >= (d + 1) * (d + 1) * (List.length sp)*) then 
      opt_starpacking rest_g sp []
    else sp
  in
(*    assert (List.fold_left (fun b s -> b && IntSet.size (snd s) = d + 2) true sp);
    assert ((IntSet.fold (fun m i -> max (List.fold_left (fun a s -> if IntSet.contains (snd s) i then a + 1 else a) 0 sp) m) (FGraph.vertices g) 0) <= 1); *)
    (*    Neato.output_graph "test.dot" (FGraph.to_graph g) 
      (List.map FGraph.to_graph (List.map (FGraph.subgraph g) (List.map snd sp)));
	  Sys.command "neato -Tps test.dot > test.ps; gv test.ps"; *)
    List.length sp
;;

(*
let rec lower_bound_f g d =
  if FGraph.max_deg g <= d then 0 else
    let rec get_vertex deg =
      let deg_vertices = FGraph.deg_vertices g deg in
	if (IntSet.is_empty deg_vertices) then
          get_vertex (deg+1)
	else
          let v, _ = IntSet.pop deg_vertices in v
    in
    let v = get_vertex (d+1) in 
    let nb = FGraph.neighbors g v in
    let rec subset s i =                         (* arbitrary size-i subset of s *)
      if i = 0 then IntSet.empty
      else
        let (el,s') = IntSet.pop s in
          IntSet.add (subset s' (i-1)) el
    in
    let nb_to_delete = subset nb (d+1) in
    let g' = IntSet.fold (fun accu i -> FGraph.delete_vertex accu i) nb_to_delete (FGraph.delete_vertex g v) in
      (lower_bound_f g' d) + 1
;;
*)

exception NO_instance;;


(* Idea: Delete high degree vertices. *)

let bddsolve_greedy_f f d =
  let rec loop f s =
    if FGraph.max_deg f <= d then s
    else 
      let v = FGraph.max_deg_vertex f in
	loop (FGraph.delete_vertex f v) (IntSet.add s v)
  in
    loop f IntSet.empty
;;

(* Idea: Packing of stars. *)
(* 

let bddsolve_greedy_f g d =
  let rec loop g s =
    if FGraph.max_deg g <= d then s
    else
      let v = FGraph.max_deg_vertex g in
      let v_n = FGraph.neighbors g v in
      let rec leaves n d g s = 
	if d < 0 then g, s
	else 
	  let leaf = IntSet.choose n in
	    leaves (IntSet.delete n leaf) (d - 1) (FGraph.delete_vertex g leaf) (IntSet.add s leaf)
      in
      let g, s = leaves v_n d (FGraph.delete_vertex g v) (IntSet.add s v) in
	loop g s
  in
    loop g IntSet.empty
;;
*)

(* kernelize with high-degree rule *)
let bddkernel_buss g d k = 
  let rec loop g d k s = 
    let s', g', k', continue = FGraph.fold_vertices (fun (saccu, gaccu, kaccu, continue) i nb -> 
						       if (FGraph.deg gaccu i) > d + kaccu then 
							 let gaccu = FGraph.delete_vertex gaccu i in
							 let saccu = IntSet.add saccu i in
							 let kaccu = kaccu - 1 in
							   if kaccu <= -1 then raise NO_instance else
							     (saccu, gaccu, kaccu, true)
						       else (saccu, gaccu, kaccu, continue)) g (s, g, k, false) in
      if (*FGraph.fold_vertices (fun accu i _ -> accu || FGraph.deg g' i > d + k') g' false*) continue then
        loop g' d k' s'
      else (s', g', k') in
    loop g d k IntSet.empty
;;

let nodecount = ref 0;;
let kernelnodecount = ref 0;;

let buss_num = ref 0;;
let buss_vertices_deleted = ref 0;;

let deg_one_num = ref 0;;
let deg_one_vertices_deleted = ref 0;;

let deg_two_num = ref 0;;
(*let deg_two_vertices_deleted = ref 0;;*)
let cliquerule_num = ref 0;;
let cliquerule_vertices_deleted = ref 0;;
let ntrule_num = ref 0;;
let ntrule_vertices_deleted = ref 0;;



type substitute_t = {
  d2 : int;
  s : int;
  r : int;
  l : int;
};;

(* Resubstitute all vertices of the degree-two-rule *)
let resubstitute s substitutes =
  (*  print_endline ("Resubstituted: " ^ string_of_int (List.length substitutes)); *)
  List.fold_left begin fun s sub ->
    if IntSet.contains s sub.s then 
      let s = IntSet.delete s sub.s in
      let s = IntSet.add s sub.r in
      let s = IntSet.add s sub.l in
	s 
    else IntSet.add s sub.d2
  end s substitutes
;;

let find_CD b d_0 p' x =
  let rec iterate v_d = 
    let v_c = 
      IntSet.intersection
	(IntSet.fold (fun accu i -> IntSet.union accu (FGraph.neighbors b i)) v_d IntSet.empty)
	x 
    in
      (*     !f_debug ("|v_c| = " ^ string_of_int(IntSet.size v_c) ^ "\n"); *)
    let v_d' = IntSet.fold (fun accu i -> if IntSet.contains (Graph.vertices p') i then (* is this "if" necessary? *)
                              IntSet.union accu (Graph.neighbors p' i) else
                                accu) v_c v_d in
      if IntSet.compare v_d' v_d = 0 then (v_c, v_d) else iterate v_d' in
    iterate d_0
;;

let maximum_star_packing b v w m = Match.maximum_edge_centered_starpacking b v w ~m ();;

let open_neighborhood g s = 
  let nb = IntSet.fold (fun accu i -> try IntSet.union accu (FGraph.neighbors g i) with Not_found -> accu) s IntSet.empty in
    IntSet.minus nb s
;; 

let closed_neighborhood g s = 
  IntSet.fold (fun accu i -> IntSet.union accu (FGraph.neighbors' g i)) s IntSet.empty 
;; 

(* Bdd-Rule *)
let bdd_nt_reduction g d greedy_solution =
(*  let dbgdotfiles = false in *)
(*  let debug str = print_string str; print_newline (); in *)

  let x = greedy_solution in
    (*  f_debug := debug; *)
  let y = IntSet.minus (FGraph.vertex_set g) x in
  let numloops = ref 0 in

  let rec loop b x y p' =
    numloops := !numloops + 1;
    let p' = maximum_star_packing b x (d+1) p' in 
      (*    !f_debug ("BDDKERNEL: Number of stars in P':  " ^ string_of_int (List.length p'));
	    if dbgdotfiles then Neato.output_graph "DEBUG-BDDKERNEL-maximum-star-packing.dot" b p'; *)
    let v_p' = Graph.vertices p' in 
    let d_0  = IntSet.minus y v_p' in
      (*  !f_debug ("BDDKERNEL: size of d_0: " ^ string_of_int(IntSet.size d_0) ^ "\n"); *)
      let (c,d) = find_CD b d_0 p' x in
      let f_y = IntSet.minus (FGraph.vertices b) (IntSet.union x d) in
	if IntSet.fold (fun accu i -> accu || IntSet.do_intersect (FGraph.neighbors g i) f_y) d false then
	  let f_y' = (IntSet.minus (closed_neighborhood g f_y) x) in
	  let f_x' = IntSet.minus x c in
	  let b, p' = 
	    IntSet.fold begin fun (b, p') i -> 
	      FGraph.delete_vertex b i, (try Graph.delete_vertex p' i with Not_found -> p')
	    end f_x' (b, p')
	  in
	  let b, p' = 
	    IntSet.fold begin fun (b, p') i -> 
	      (try FGraph.delete_vertex b i with Not_found -> b), (try Graph.delete_vertex p' i with Not_found -> p')
	    end f_y' (b, p')
	  in
	  let x' = c in
	  let y' = IntSet.minus y f_y' in
	    loop b x' y' p'
	else (c,d) 
  in (* loop *)

  let (c,d) = loop g x y Match.matching_empty in
  (*
  Printf.printf "%s" ("BDDKERNEL: size of c: " ^ string_of_int(IntSet.size c) ^ "\n");
  Printf.printf "%s" ("BDDKERNEL: size of d: " ^ string_of_int(IntSet.size d) ^ "\n"); 
  Printf.printf "%s" ("BDDKERNEL: number of iterations: " ^ string_of_int(!numloops) ^ "\n");
  *)

    c, d
;;

(* ----- Bdd search tree algorithm ----- *)

(* Define some measures *)
let nodecount = ref 0;;
let bounded_degree_nodecount = ref 0;;
let finished_after_simple_interleaving_nodecount = ref 0;;
let no_instance_after_simple_interleaving_nodecount = ref 0;;
let bound_rule_nodecount = ref 0;;
let edge_rule_nodecount = ref 0;;
let forbidden_nodecount = ref 0;;
let no_instance_due_forbidden_vertices_nodecount = ref 0;; 
let no_instance_due_branching_nodecount = ref 0;; 
let rrules_time = ref 0.0;;
let ntrule_time = ref 0.0;;
let branch_vertex_selection_time = ref 0.0;;
let greedy_sol_time = ref 0.0;;
let preprocess_time = ref 0.0;;
let edge_rule_time = ref 0.0;;
let lb_rule_time = ref 0.0;;

(* Fold all subsets of given size of an IntSet *)
let fold_subsets f (accu: bool * IntSet.t) intset size =
  let rec loop intset' subset accu =
    if IntSet.size intset' < size - IntSet.size subset then
      accu
    else if IntSet.size intset' = size - IntSet.size subset then
      f accu (IntSet.union intset' subset)
    else if IntSet.size subset = size then 
      f accu subset
    else
      let elem, intset' = IntSet.pop intset' in
      let accu = loop intset' (IntSet.add subset elem) accu in
	loop intset' subset accu in
    loop intset IntSet.empty accu
;;

(* Choose a vertex to branch on, return it and its neighbors *)
let branch_vertex_selection d g' greedy_solution forbidden_vertices =
  let v, nb =
    if not (!guided_branch_vertex_selection || !forbid_branch_vertices) then
    let v = FGraph.max_deg_vertex g' in
      v, FGraph.neighbors g' v
    else
    (* Choose branch vertex candidates *)
    let candidates = 

      (* Ensure enough candidates are available *)
      let c, unused_degrees = 
	let rec min_candidate_loop c degrees =
	  if IntSet.size c > !min_branch_candidates then c, degrees
	  else if IntMap.is_empty degrees || IntMap.max_key degrees <= d 
	  then 
	    if IntSet.size c > 0 then c, degrees
	    else raise NO_instance
	  else
	    let m = IntMap.max_key degrees in
	    let m_v = IntMap.get degrees m in
	    let degrees = IntMap.delete degrees m in
	    let c = IntSet.union c m_v in
	      min_candidate_loop c degrees
	in
	  min_candidate_loop (FGraph.max_deg_vertices g') (FGraph.deg_struct g')
      in
	
	(* If any candidates are forbidden, search for new ones. *)
	(* maybe TODO: upper bound for number of candidates. *)
	if !forbid_branch_vertices && d > 0 then
	  let filter_forbidden c = IntSet.minus c forbidden_vertices in
	  let rec forbid_loop c degrees =
	    if IntSet.size c >= !min_branch_candidates then c
	    else if IntMap.is_empty degrees || IntMap.max_key degrees <= d then
	      if IntSet.size c > 0 then c
	      else raise NO_instance
	    else 
	      let m = IntMap.max_key degrees in
	      let m_v = IntMap.get degrees m in
	      let m_v = filter_forbidden m_v in
	      let degrees = IntMap.delete degrees m in
	      let c = IntSet.union c m_v in
		forbid_loop c degrees
	  in
	    forbid_loop (filter_forbidden c) unused_degrees
	else c
    in

      (* Search for a candidate with maximum number of neighbors in the greedy solution. *)
      if !guided_branch_vertex_selection then 
	let nb_greedy_intersect w =
	  let nb_w = FGraph.neighbors g' w in
	  let n_w = IntSet.size (IntSet.intersection (IntSet.add nb_w w) greedy_solution) in
	    n_w, nb_w
	in
	  
	let fold (v, n, nb) w =
	  let n_w, nb_w = nb_greedy_intersect w in
	    if n_w > n then (w, n_w, nb_w) else (v, n, nb)
	in
	  
	let v = IntSet.choose candidates in
	let n_v, nb_v = nb_greedy_intersect v in
	let v, _, nb = IntSet.fold fold (IntSet.delete candidates v) (v, n_v, nb_v) in
	  v, nb
      else 
	let v = IntSet.choose candidates in 
	  v, FGraph.neighbors g' v
  in
    v, nb
;;


exception Time_out;;

(* Forbidding solution does not yet work as fast as necessary to be efficient.*)
let forbidden_solutions = ref [];;

(*
let partial_solution_forbidden status =
  let solution_forbidden = 
    !forbid_sol_subsets &&
      List.fold_left begin fun solution_forbidden sol ->
	solution_forbidden || IntSet.is_subset sol status.s
      end false !forbidden_solutions
  in
    if solution_forbidden then forbidden_nodecount := !forbidden_nodecount + 1;
    solution_forbidden
;;

*)

let rec bddsolve' g d k time_available =
  let timer_start = Util.timer () in

  let time_out _ = 
    time_available >= 0.0 && (Util.timer ()) -. timer_start > time_available 
  in

  let max_degree_satisfied g = FGraph.max_deg g <= d in

  let return solution_found s substitutes = 
    if solution_found then
      let s =
	if !deg_two_interleaving && d = 0 then 
	  resubstitute s substitutes (* TODO Is recursion faster than 
					carrying the substitutes around? *)
	else s
      in
	true, s
    else
      false, IntSet.empty
  in

  let solution_too_big k = k < 0 in

  let solution_found = fst in

  (* Add vertices to the solution and remove them from the instance. *)
  let delete g k s greedy_solution new_sol_vertices =
    let g = IntSet.fold FGraph.delete_vertex new_sol_vertices g 
    and greedy_solution = IntSet.minus greedy_solution new_sol_vertices 
    and s = IntSet.union s new_sol_vertices 
    and k = k - (IntSet.size new_sol_vertices) 
    in
      g, k, s, greedy_solution
  in

  (* Remove needless vertices. *)
  let remove g greedy_solution no_sol_vertices =
    let g = IntSet.fold FGraph.delete_vertex no_sol_vertices g
    and greedy_solution = IntSet.minus greedy_solution no_sol_vertices 
    in
      g, greedy_solution
  in

  let recompute_greedy_solution g greedy_solution level = (* TODO multiple greedy solvers. *)
    if !greedyopt_interval > 0 && 
      (level mod (!greedyopt_interval + 1)) = !greedyopt_interval 
    then
      let greedy_solution' = bddsolve_greedy_f g d in 
	if IntSet.size greedy_solution' < IntSet.size greedy_solution then
	  greedy_solution'
	else greedy_solution
    else
      greedy_solution
  in

  let greedy_solvable k greedy_solution = k >= IntSet.size greedy_solution in

  (* RRR: Reduction and removal rules. *)
(*  let increment counter_ref incr = counter_ref := !counter_ref + incr in
    
  let count *)


  (* RRR: High degree rule. *)
  let buss g k s greedy_solution =
    let to_delete = FGraph.deg_rel g ((<) (d + k)) in
    let g, k, s, greedy_solution = 
      delete g k s greedy_solution to_delete
    in 
      g, k, s, greedy_solution, IntSet.size to_delete, 0
  in

  (* RRR: Remove vertices of degree at most d whose adjacent vertices have also
     degree at most d. *)
  let deg_leq_d g k s greedy_solution = 
    let leq_d = FGraph.deg_rel g ((>=) d) in
    let del = IntSet.fold begin fun del v ->
      let n = FGraph.neighbors g v in
	if IntSet.is_subset n leq_d then IntSet.add del v else del
    end leq_d IntSet.empty
    in
    let g, greedy_solution = remove g greedy_solution del in
      g, k, s, greedy_solution, 0, IntSet.size del
  in

  (* RRR: Degree-one rule. *)
  let deg_one g k s greedy_solution =
    (* Deletes both the vertex and its now zero-degree neighbors. *)
    let fold = fun (accu, num) v n ->
      if IntSet.size n = 1 then
	let n = IntSet.choose n in
	let num = 
	  IntMap.modify_default begin fun (no, d1n) -> 
	    succ no, IntSet.add d1n v 
	  end num n (0, IntSet.empty) 
	in
	let accu = 
	  let no, d1n = IntMap.get num n in
	    if no > d then 
	      IntMap.modify_default 
		begin fun x -> d1n end 
		accu n d1n 
	    else accu 
	in
	  accu, num
      else
	accu, num
    in
    let accu, _ = 
      FGraph.fold_vertices fold g (IntMap.empty, IntMap.empty)
    in
      (* Two incident degree-one vertices (d=0) are deleted 'first come,
	 first serve' (hence the possible Not_found exception). *)
      IntMap.fold begin fun 
	((g, k, s, greedy_solution, deleted, removed) as accu) n d1n ->
	  if FGraph.has_vertex g n then
	    let g, greedy_solution = remove g greedy_solution d1n in
	    let g, k, s, greedy_solution = 
	      delete g k s greedy_solution (IntSet.singleton n)
	    in
	      g, k, s, greedy_solution, deleted + 1, removed + (IntSet.size d1n)
	else accu
      end accu (g, k, s, greedy_solution, 0, 0)
  in

  (* RRR: Degree-two rule for d=0. *)
  let deg_two g k s greedy_solution substitutes =
    IntSet.fold 
      begin fun ((g, k, s, greedy_solution, 
		  substitutes, deleted, removed) as accu) i ->
	let n = try FGraph.neighbors g i with Not_found -> IntSet.empty in
	  if IntSet.size n <> 2 then accu
	  else 
	    let r = IntSet.min n and l = IntSet.max n in
	    let r_n = IntSet.delete (FGraph.neighbors g r) i in
	      deg_two_num := !deg_two_num + 1; (* TODO normalize counting. *)
	      if IntSet.contains r_n l then
		let g, k, s, greedy_solution = 
		  delete g k s greedy_solution n 
		in
		let g, greedy_solution = 
		  remove g greedy_solution (IntSet.singleton i)
		in
		  g, k, s, greedy_solution, 
	            substitutes, deleted + 2, removed + 1
	      else     
		let l_n = IntSet.delete (FGraph.neighbors g l) i in
		let g = IntSet.fold (FGraph.delete_vertex) n g in
		let connector = begin fun g m -> 
		  try FGraph.connect g i m with IntSet.Already_present -> g 
		end in
		let g = IntSet.fold connector r_n g in
		let g = IntSet.fold connector l_n g in
		let substitutes = 
		  {d2 = i; s = i; r = r; l = l} :: substitutes 
		in
		let greedy_solution =
		  if IntSet.do_intersect greedy_solution n then 
		    IntSet.put (IntSet.minus greedy_solution n) i
		  else IntSet.minus greedy_solution (IntSet.add n i)
		in
		  g, k - 1, s, greedy_solution, 
	            substitutes, deleted, removed + 2
      end (FGraph.deg_vertices g 2) 
      (g, k, s, greedy_solution, substitutes, 0, 0)
  in


  (* RRR: Rule to remove some small cliques for d = 0 *)
  let cliquerule g k s greedy_solution =
    let g', s' = IntSet.fold 
      (fun (gaccu,saccu) i ->
         if FGraph.has_vertex gaccu i && FGraph.deg gaccu i <= 5 then
           let neighbors = FGraph.neighbors gaccu i in
           let gaccu, _, saccu = IntSet.fold
             (fun (gaccu', nbaccu, saccu') j ->
                if IntSet.size (IntSet.intersection nbaccu (FGraph.neighbors gaccu j)) = ((IntSet.size nbaccu) - 1) then
                  ((FGraph.delete_vertex gaccu' j), (IntSet.delete nbaccu j), (IntSet.add saccu' j))
                else
                  (gaccu', nbaccu, saccu')
             ) neighbors (gaccu, neighbors, saccu) in
             (gaccu, saccu)
         else
           (gaccu, saccu)
      ) (FGraph.vertices g) (g, IntSet.empty) in
      (* TODO measuring *)
      cliquerule_num := !cliquerule_num + (if not (IntSet.is_empty s) then 1 else 0);
      cliquerule_vertices_deleted := !cliquerule_vertices_deleted + (IntSet.size s);
      let g, k, s, greedy_solution = delete g k s greedy_solution s' in
	g, k, s, greedy_solution, IntSet.size s', 0
  in

  (* RRR: BDD-Rule *)
  let bdd_rule g k s greedy_solution = 
    let x = IntSet.size greedy_solution in
    let kernelizable = !nt_kernelization 
      && IntSet.size (open_neighborhood g greedy_solution) > x * (d + 1)
    in
      
      if kernelizable then
        let c, d = bdd_nt_reduction g d greedy_solution in
	let g, k, s, greedy_solution = 
	  delete g k s greedy_solution c 
	in
	let g, greedy_solution = remove g greedy_solution d in
	  g, k, s, greedy_solution, IntSet.size c, IntSet.size d
      else g, k, s, greedy_solution, 0, 0
  in

  (* RRR: Main Loop. *)
  let interleaving_reduction g k s greedy_solution substitutes =
      
      (* TODO measuring *) (* TODO do this beautifully *)
    let rec reduct_loop g k s greedy_solution substitutes =
      let g, k, s, greedy_solution, buss_deleted, _ = 
	if !buss_interleaving then buss g k s greedy_solution  
	else g, k, s, greedy_solution, 0, 0
      in
      let g, k, s, greedy_solution, deg_one_deleted, _ = 
	if !deg_one_interleaving then deg_one g k s greedy_solution
	else g, k, s, greedy_solution, 0, 0
      in
      let g, k, s, greedy_solution, substitutes, 
	deg_two_deleted, deg_two_removed = 
	if !deg_two_interleaving && d = 0 then 
	  deg_two g k s greedy_solution substitutes
	else g, k, s, greedy_solution, substitutes, 0, 0
      in
      let g, k, s, greedy_solution, clique_deleted, _ =
	if !cliquerule_interleaving && d = 0 then 
	  cliquerule g k s greedy_solution
	else g, k, s, greedy_solution, 0, 0
      in
      let g, k, s, greedy_solution, bdd_rule_deleted, _ = 
	if !nt_kernelization then bdd_rule g k s greedy_solution
	else g, k, s, greedy_solution, 0, 0
      in
	
	if buss_deleted > 0 || deg_one_deleted > 0 
	  || bdd_rule_deleted > 0 || deg_two_deleted > 0
	  || deg_two_removed > 0 || clique_deleted > 0
	then reduct_loop g k s greedy_solution substitutes
	else g, k, s, greedy_solution, substitutes
    in
      
      reduct_loop g k s greedy_solution substitutes
  in
  (* RRR: End. *)

  (* Lower bound rule; *)
  let lower_bound_rule g k =
    !bound_rule &&
      let lower_bound = lower_bound_f g d in
	k < lower_bound
  in

  (* Edge rule. *)
  let edge_rule g k =
    let rec deletable_loop k ds edges =
      if IntMap.is_empty ds then edges
      else 
	let m = IntMap.max_key ds in
	  if m <= d then edges
	  else
	    let n = IntSet.size (IntMap.get ds m) in
	    let ds = IntMap.delete ds m in
	      if n > k then edges + k * m
	      else deletable_loop (k - n) ds (edges + n * m)
    in    
      !edge_rule &&
	let deletable_edges = 
	  deletable_loop k (FGraph.deg_struct g) 0 
	in
	  (FGraph.num_edges g - deletable_edges) 
	  > (d * (FGraph.num_vertices g) / 2)
  in


  (* Search tree main loop. *)	 
  let rec search_loop g k s greedy_solution level substitutes 
      forbidden_vertices path_string =
    
    (* Cut: Search for small vertex sets, that separate the graph. *)
    let search_cut g =
      let cut_candidates = Cut.cut_corner (FGraph.to_graph g) in
	List.fold_left begin fun (best_cut, components) (s, _) ->
       (* TODO better cut-selection? *)
	  if not (IntSet.is_empty best_cut) then
	    (best_cut, components)
	  else
	    let g = IntSet.fold (FGraph.delete_vertex) s g in
	    let s_components = Cut.connected_components (FGraph.to_graph g) in
	      if List.length s_components > 1 then
		(s, s_components)
	      else (best_cut, components)
	end (IntSet.empty, []) cut_candidates 
    in

    (* Cut: Try to reduce the instance through solving subgraphs induced
       by cut-sets and adjacent connected components. *)
    let use_cut g k s greedy_solution cut components =
  (*     prerr_endline "Found cut."; *)
     (*
      let upper_bounds, _ = 
	List.fold_left begin fun (upper_bounds, component_no) component ->
	  let upper_bound = 
	    IntSet.size (bddsolve_greedy_f (FGraph.subgraph g component) d)
	  in 
	  let upper_bounds =
	    IntMap.modify_default (fun x -> IntSet.add x component_no)
	      upper_bounds upper_bound IntSet.empty
	  in
	    upper_bounds, component_no + 1
	end (IntMap.empty, 0) components 
      in
      let max_k = 
	IntMap.fold (fun i j s -> i + (j * (IntSet.size s))) upper_bounds 0 
      in
      let max_k = max_k - (IntMap.max_key upper_bounds) in
      let component = 
	IntSet.choose (IntMap.get upper_bounds (IntMap.max_key upper_bounds))
      in
      let component = List.nth components component in
      let g_small = FGraph.subgraph g component in
      let max_k_small = k - max_k - (IntSet.size cut) in
	if max_k_small < 0 then begin
	  prerr_endline "Not easy to cut.";
	  g, k, s, greedy_solution
	end else
	  let greedy_solution_small = bddsolve_greedy_f g_small d in
	    if (IntSet.size greedy_solution_small) 
	      + (IntSet.size cut)
	      + max_k < k 
	    then begin
	      (* prerr_endline "Greedy cut reduction solved the instance."; *)
	      (* TODO return new instance with valid greedy solution *)
	      g, k, s, greedy_solution
	    end else
	      let solution_found, solution = 
		search_loop g_small max_k_small IntSet.empty 
		  greedy_solution_small level []
		  forbidden_vertices path_string
	      in
		if not solution_found then begin
		  prerr_endline "Found no cut reduction.";
		  g, k, s, greedy_solution
		end else begin
   (*		  Printf.eprintf "Found cut reduction. Component size: %i; Cut size %i\n" (IntSet.size cut) (IntSet.size component); *)
		  delete g k s greedy_solution (IntSet.union solution cut)
		end *)
      
      let smallest_component =
	List.fold_left begin fun smallest_component component ->
	  if IntSet.size smallest_component > IntSet.size component 
	  then component 
	  else smallest_component
	end (List.hd components) (List.tl components)
      in
      let g_small = FGraph.subgraph g (IntSet.union smallest_component cut) in
      let solution1, _, _ = bddsolve_optimal g_small d (-1.0) in
      let solution2, _, _ = 
	bddsolve_optimal (FGraph.subgraph g_small smallest_component) d (-1.0)
      in
	if (IntSet.size solution2) + (IntSet.size cut) <= (IntSet.size solution1)
	then begin	
(*	  Printf.eprintf "Found cut reduction. Component size: %i; Cut size %i\n" (IntSet.size cut) (IntSet.size smallest_component); *)
	  delete g k s greedy_solution (IntSet.union solution2 cut)
	end else g, k, s, greedy_solution
    in
      
    (* Branch: Search a branching vertex and branch into all necessary 
       cases. *)
    let branch g k s greedy_solution substitutes =
      (* Branch: Low-level branching function for deleting sets of vertices. *)
      let branch_on_vset path_string forbidden_vertices vset =
	let g, k, s, greedy_solution = delete g k s greedy_solution vset in
	  if solution_too_big k then return false s substitutes
	  else search_loop g k s greedy_solution level substitutes
	    forbidden_vertices path_string
      in
	
      (* Branch: Delete the branching vertex. *)
      let branch_on_vertex v = 
	branch_on_vset (path_string ^ "s") 
	  forbidden_vertices (IntSet.singleton v)
      in

      (* Branch: Handle the returning branches. *)
      let handle_branch_vl value1 branch2 =
	let (solution_found, _) = value1 in
	  if solution_found then value1
	  else Lazy.force branch2
      in
      let handle_branch_ll branch1 branch2 =
	handle_branch_vl (Lazy.force branch1) branch2
      in (* TODO: Do the handling directly? *)
	
      (* Branch on neighbors: Try all necessary subsets of the set of 
	 neighbor-vertices. *)
      let branch_on_neighbors branch_vertex neighbors =

	(* Branch on neighbors: Forbid the branching vertex in further 
	   branching. *)
	let forbidden_vertices =
	  if !forbid_branch_vertices && d > 0 then
	    IntSet.add forbidden_vertices branch_vertex
	  else forbidden_vertices
	in
	let path_string = path_string ^ "n" in

	(* Branch on neighbors: Regular branching. *)
	let branch_regular neighbors =
	  let branch_size = IntSet.size neighbors - d in
	    fold_subsets begin fun return_value vset -> 
	      if solution_found return_value then return_value
	      else
		(branch_on_vset path_string 
		   forbidden_vertices vset)
	    end (false, IntSet.empty) neighbors branch_size
	in
	  
	(* Branch on neighbors: Greedybranching functions. *)
	let greedysol_branching_condition neighbors greedy_neighbors =	
	  (IntSet.size greedy_neighbors > d) 
	  && ((float(IntSet.size greedy_neighbors) 
	       /. float(IntSet.size neighbors)) > 0.9 )
	in

	let branch_in_greedy_solution = branch_regular in 
	  (* Only arguments differ.*)
	  
	(* Branch on neighbors: Improved branching for d=1 *)
	let branch_d1 neighbors =
	  let branch_size = IntSet.size neighbors - d in
	  let return_value = 
	    branch_on_vset path_string forbidden_vertices neighbors 
	  in
	    fold_subsets begin fun return_value vset ->
	      if solution_found return_value then return_value
	      else
		let remaining_vertex = 
		  IntSet.choose (IntSet.minus neighbors vset)
		in
		let neighbors_neighbors = 
		  IntSet.delete 
		    (FGraph.neighbors g remaining_vertex)
		    branch_vertex
		in
		  branch_on_vset path_string forbidden_vertices 
		    (IntSet.union vset neighbors_neighbors)
	    end return_value neighbors branch_size
	in
	  
	  (* Branch on neighbors: High-level neighbor branching. *)
	  if !greedysol_branching then
	    let greedy_neighbors = 
	      IntSet.intersection neighbors greedy_solution 
	    in
	      if greedysol_branching_condition neighbors greedy_neighbors 
	      then 
		branch_in_greedy_solution greedy_neighbors
	      else branch_regular neighbors (* TODO d=1 then branch_d1 *)
	  else if d = 1 then 
	    branch_d1 neighbors
	  else
	    branch_regular neighbors
      in

      (* Branch: Select branching vertex. *)
      let select_branch_vertex =
	try 	(* TODO: exception *)
	  let ret = branch_vertex_selection d g greedy_solution 
	    forbidden_vertices
	  in
	    ret, true 
	with NO_instance -> (0, IntSet.empty), false
      in

      (* TODO: greedy_neighbors computed two times. *)
      let (branch_vertex, neighbors), continue = 
	select_branch_vertex
      in
	if not continue then return false s substitutes
	else
	
	  (* Branch: Depending on membership in the greedy solution, branch first 
	     on the branching vertex or its neighbors. *)
	  if !greedysol_branching
	    && IntSet.contains greedy_solution branch_vertex 
	  then
	    let return_value = branch_on_vertex branch_vertex in
	      if solution_found return_value then return_value
	      else branch_on_neighbors branch_vertex neighbors
	  else
	    let return_value = branch_on_neighbors branch_vertex neighbors in
	      if solution_found return_value then return_value
	      else branch_on_vertex branch_vertex
	      (* TODO forbidden solution subsets. *)
    in
    (* Branch: End. *)

    (* High-level search tree. *)
    if time_out () then raise Time_out
    else 
      incr nodecount;
      if max_degree_satisfied g then return true s substitutes
	(*    else if partial_solution_forbidden status then return false status  TODO *)
      else 
	let g, k, s, greedy_solution, substitutes = 
	  interleaving_reduction g k s greedy_solution substitutes 
	in 
	  if solution_too_big k then return false s substitutes
	  else if max_degree_satisfied g then return true s substitutes
	  else 
	    if lower_bound_rule g k then return false s substitutes
	    else if edge_rule g k then return false s substitutes
	    else 
	      let greedy_solution = 
		recompute_greedy_solution g greedy_solution level
	      in
		if greedy_solvable k greedy_solution then 
		  let s = IntSet.union s greedy_solution in
		    return true s substitutes
		else
		  (*  let cut, components = search_cut g in
		      let g, k, s, greedy_solution =
		      if List.length components > 1 then
		      use_cut g k s greedy_solution cut components
		      else g, k, s, greedy_solution
		      in  *)
		  branch g k s greedy_solution substitutes
  in
    
  let solution_found, solution =
    search_loop g k IntSet.empty (bddsolve_greedy_f g d) 0 [] IntSet.empty ""
  in
    if solution_found then solution else raise NO_instance

and bddsolve_optimal g d time_available =
  (* TODO timers *)
  let greedy_solution = bddsolve_greedy_f g d in
  let upper_bound = IntSet.size greedy_solution in
  let lower_bound = lower_bound_f g d in
    
  let rec loop g k sol time_available =
    if k >= lower_bound then begin
      (* TODO if !verbose then debug *) 
(*      Printf.eprintf "Trying k = %i\n" k; *)
      flush stderr;
      let continue, time_out, sol =
	try
	  true, false, bddsolve' g d k (-1.0)
	with
	    NO_instance -> false, false, sol
	  | Time_out -> false, true, sol
      in
	if not continue then
	  if time_out then begin
	    (* TODO if !verbose then debug *) Printf.eprintf "Time out.\n";
	    sol, lower_bound, k + 1
	  end else
	    sol, IntSet.size sol, IntSet.size sol
	else
	  loop g ((IntSet.size sol) - 1) sol time_available
    end else sol, IntSet.size sol, IntSet.size sol
  in 
    loop g (upper_bound - 1) greedy_solution time_available
;;

(* Compression step of Iterative Compression (not implemented) *)
(*
let itco_compression g s =
  s
;;
*)

(* Solve d-BDD via Iterative Compression. (For testing purposes only! This
  function is not completely implemented!) *)
(*
let bddsolve_itco g d =
  let s = bddsolve_greedy g d in
  let g' = FGraph.subgraph g s in

(* Iterative compression outer loop (finding all subsets of the solution s
   for which we have to apply the compression step *)
  let fold_bdd_subsets f accu g d = 
    let rec loop f accu g bddset = 
      if FGraph.num_vertices g = 0 then (* Every vertex has been processed *)
        f accu bddset
      else
        if d <= 1 then
          let v = FGraph.max_vertex g in 
          let g' = FGraph.delete_vertex g v in
          let accu = loop f accu g' bddset in
          let bddset = IntSet.add bddset v in
          let nb = FGraph.neighbors g v in
          let g' = FGraph.delete_vertices g' nb in
          let accu = loop f accu g' bddset in
            if d = 1 then
              IntSet.fold 
		(fun accu' i -> 
                   let vset = IntSet.minus (FGraph.neighbors g i) (IntSet.add nb v) in
                   let g' = FGraph.delete_vertices g' vset in
                   let bddset = IntSet.add bddset i in
                     loop f accu' g' bddset)
		nb accu
            else accu
        else
          begin
            Printf.eprintf "ERROR: Itco with d >= 2 is not supported\n";
            exit 1;
            accu
          end
    in
      loop f accu g IntSet.empty
  in
  let test = fold_bdd_subsets (fun accu solution -> 
                                 accu
			      ) 1000000 g' d in
    Printf.eprintf "LB: %i" test;
    IntSet.empty
;;
*)
